package info.kwarc.mmt.frameit.rules

import info.kwarc.mmt.api.checking._
import info.kwarc.mmt.api.{DPath, GlobalName, LocalName, ParametricRule, Rule, RuleSet}
import info.kwarc.mmt.api.frontend.Controller
import info.kwarc.mmt.api.objects._
import Conversions._
import info.kwarc.mmt.api.uom.{Simplifiability, Simplify}
import info.kwarc.mmt.lf.{Apply, ApplySpine, Arrow, OfType, Pi}
import info.kwarc.mmt.LFX.datatypes.{Listtype, Ls}
import info.kwarc.mmt.api.utils.URI

import scala.collection.mutable

object StepUntil {
  val ns: DPath = DPath(URI.http colon "mathhub.info") / "FrameIT" / "frameworld"
  val path: GlobalName = ns ? "StepUntil" ? "stepUntil"
  val rawhead: OMID = OMS(path)
  def apply(tp1: Term, iv: Term, f: Term, tC: Term): Term =
    ApplySpine(this.rawhead, tp1, iv, f, tC)
  def unapply(tm : Term) : Option[(Term, Term, Term, Term)] = tm match {
    case ApplySpine(this.rawhead, List(tp1, iv, f, tC)) => Some((tp1, iv, f, tC))
    case _ => None
  }
}

object StepUntilRule extends ParametricRule {
  // todo: document
  private val DEBUG_MAX_ITER_COUNT = 100

  override def apply(controller: Controller, home: Term, args: List[Term]): Rule = args match {
    case List(truetm) =>
      RuleSet(StepUntilComp(truetm))
    case _ =>
      ???
  }

  case class StepUntilComp(truetm: Term) extends ComputationRule(StepUntil.path) {
    override def applicable(t: Term): Boolean = t match {
      case StepUntil(_, _, _, _) => true
      case _ => false
    }
    override def apply(check: CheckingCallback)(tm: Term, covered: Boolean)(implicit stack: Stack, history: History): Simplifiability = tm match {
      case StepUntil(_, iv, f, tC) =>
        val values = mutable.ListBuffer[Term]()
        var iter = 0
        var cur = iv
        while (iter < DEBUG_MAX_ITER_COUNT) {
          iter += 1
          check.tryToCheckWithoutDelay(Equality(stack, Apply(tC, cur), truetm, None)) match {
            case Some(true) =>
              return Simplify(Ls(values.toSeq : _*))
            case Some(false) | None =>
              values += cur
              cur = check.simplify(Apply(f, cur))
          }
        }
        // todo: try to log something via stack, history, check, or this
        Simplify(Ls(values.toSeq : _*))
      case _ => Simplifiability.NoRecurse
    }
  }
}